(*
 * This file is part of Barista.
 * Copyright (C) 2007-2012 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

let register = Camlp4.PreCast.AstFilters.register_str_item_filter

module Make (Syntax : Camlp4.Sig.Camlp4Syntax) = struct
  open Camlp4.Sig
  include Syntax

  (* Shorthand for the 'unknown' location. *)
  let ghost = Loc.ghost


  (* Symbolic constants for the UTF8-related functions. *)
  let utf8_of_string = <:expr@ghost< Utils.UTF8.of_string >>

  let utf8_equal = <:expr@ghost< Utils.UTF8.equal >>

  let uchar_of_char = <:expr@ghost< Utils.UChar.of_char >>

  let uchar_equal = <:expr@ghost< Utils.UChar.equal >>


  (* Construction of containers based on hash tables. *)
  let create p f =
    let prefix = String.copy p in
    let prefix x = prefix ^ (string_of_int x) in
    let table = Hashtbl.create 17 in
    let add x =
      let id =
        try
          Hashtbl.find table x
        with Not_found ->
          let len = Hashtbl.length table in
          Hashtbl.add table x len;
          len in
      prefix id in
    let prepend x =
      Hashtbl.fold
        (fun k v acc ->
          let open Ast in
          let _loc = Loc.ghost in
          let id = prefix v in
          let v = f k in 
          let p = PaId (_loc, (IdLid (_loc, id))) in
          let e = StVal (_loc, ReNil, BiEq (_loc, p, v)) in
          StSem (_loc, e, acc))
        table
        x in
    add, prepend


  (* Containers for UTF8 strings and characters. *)
  let add_string, prepend_strings =
    create
      "__bartista__utf8__string__"
      (fun x ->
        let e = Ast.ExStr (ghost, x) in
        <:expr@ghost< $utf8_of_string$ $e$ >>)

  let add_char, prepend_chars =
    create
      "__bartista__uchar__"
      (fun x ->
        let e = Ast.ExChr (ghost, x) in
        <:expr@ghost< $uchar_of_char$ $e$ >>)


  (* Utility functions for construction of type/pattern tuples. *)
  type ty_sep = TyStar | TyOr

  let ty_tuple sep = function
    | [] -> Ast.TyNil ghost
    | hd :: [] -> hd
    | hd :: tl ->
        List.fold_left
          (fun acc elem ->
            match sep with
            | TyStar -> Ast.TySta (ghost, acc, elem)
            | TyOr -> Ast.TyOr (ghost, acc, elem))
          hd
          tl

  let pa_tuple cstr = function
    | [] -> <:patt@ghost< $uid:cstr$ >>
    | hd :: [] -> <:patt@ghost< $uid:cstr$ $hd$ >>
    | hd :: tl ->
        let res =
          List.fold_left
            (fun acc elem -> Ast.PaCom (ghost, acc, elem))
            hd
            tl in
        let res = Ast.PaTup (ghost, res) in
        <:patt@ghost< $uid:cstr$ $res$ >>


  (* Instrument to add UTF8 constants declaration at file head. *)
  let instrumented = ref []
  let () =
    let instrument =
      object
        inherit Ast.map as super
        method! str_item si =
          let loc = Ast.loc_of_str_item si in
          let file = Loc.file_name loc in
          if not (List.mem file !instrumented) then begin
            instrumented := file :: !instrumented;
            prepend_strings (prepend_chars si)
          end else
            si
      end in
    register (Obj.magic (instrument#str_item))


  (* Actual grammar extension. *)
  EXTEND Gram
    GLOBAL: sig_item str_item expr match_case0;
    (* "Exception" pattern. *)
    sig_error_kind: [[
      id = UIDENT; l = OPT [ "of"; t = ctyp -> t] ->
        match l with
        | Some (Ast.TyTup (_, l)) ->
            let rec unroll = function
              | Ast.TySta (_, left, right) -> (unroll left) @ (unroll right)
              | x -> [x] in
            (id, unroll l)
        | Some x -> (id, [x])
        | None -> (id, [])
    ]];
    sig_item: [[
      "BARISTA_ERROR"; "="; OPT "|"; l = LIST1 sig_error_kind SEP "|" ->
        let kinds_for_typ =
          List.map
            (fun (cstr, params) ->
              let params = ty_tuple TyStar params in
              <:ctyp< $uid:cstr$ of $params$ >>)
            l in
        let typ = ty_tuple TyOr kinds_for_typ in
        <:sig_item< type error = $typ$;;
                     exception Exception of error;;
                     val string_of_error : error -> string;; >>
    ]];
    str_error_elem: [[
      "("; id = LIDENT; ":"; typ = ctyp; ")" ->
        (id, typ)
    ]];
    str_error_kind: [[
      id = UIDENT; l = OPT [ "of"; l = LIST1 str_error_elem SEP "*" -> l]; "->"; e = expr ->
        match l with
        | Some l -> (id, l, e)
        | None -> (id, [], e)
    ]];
    str_item: [[
      "BARISTA_ERROR"; "="; OPT "|"; l = LIST1 str_error_kind SEP "|" ->
        let kinds_for_typ =
          List.map
            (fun (cstr, params, _) ->
              let params = ty_tuple TyStar (List.map snd params) in
              <:ctyp< $uid:cstr$ of $params$ >>)
            l in
        let typ = ty_tuple TyOr kinds_for_typ in
        let kinds_for_pat =
          List.map
            (fun (cstr, params, exp) ->
              let patt =
                pa_tuple cstr
                  (List.map (fun (x, _) -> <:patt< $lid:x$ >>) params) in
              Ast.McArr (Loc.ghost, patt, <:expr< >>, exp))
            l in
        let cases = <:match_case< $list:kinds_for_pat$ >> in
        <:str_item< type error = $typ$;;
                    exception Exception of error;;
                    let fail e = raise (Exception e);;
                    let string_of_error e = match e with $cases$;;
                    let () =
                      Printexc.register_printer
                        (function
                          | Exception e -> Some (string_of_error e)
                          | _ -> None) >>
    ]];
    (* UTF8 extensions. *)
    expr: LEVEL "simple" [
      [ "@"; s = STRING ->
        let id = add_string s in
        <:expr< $lid:id$ >> ]
      | [ "@"; c = CHAR ->
        let id = add_char c in
        <:expr< $lid:id$ >> ]
    ];
    match_case0: [
      [ "@"; s = STRING; "->"; e = expr ->
        let id = add_string s in
        let id' = id ^ "'" in
        <:match_case< $lid:id'$ when $utf8_equal$ $lid:id$ $lid:id'$ -> $e$ >> ]
      | [ "@"; s = STRING; "when"; w = expr; "->"; e = expr ->
          let id = add_string s in
          let id' = id ^ "'" in
          <:match_case< $lid:id'$ when ($utf8_equal$ $lid:id$ $lid:id'$) && $w$ -> $e$ >> ]
      | [ "@"; c = CHAR; "->"; e = expr ->
          let id = add_char c in
          let id' = id ^ "'" in
          <:match_case< $lid:id'$ when $uchar_equal$ $lid:id$ $lid:id'$ -> $e$ >> ]
      | [ "@"; c = CHAR; "when"; w = expr; "->"; e = expr ->
          let id = add_char c in
          let id' = id ^ "'" in
          <:match_case< $lid:id'$ when ($uchar_equal$ $lid:id$ $lid:id'$) && $w$ -> $e$ >> ]
    ];
  END
end


(* Actual registration of syntax extension and instrument. *)
let () =
  let open Camlp4 in
  let module Id =
    struct
      let name = "Barista-syntax-extension"
      let version = ""
    end in
  let module M = Camlp4.Register.OCamlSyntaxExtension (Id) (Make) in
  ()
